/*
 * uvector - uniform (homogeneous) numeric vector types (SRFI-4)
 *
 *  (c) 1999 Shiro Kawai  shiro@acm.org
 *  Terms and conditions to use, modify or distribute this code
 *  is the same as of STk.
 *
 *  $Id: uvector.h,v 1.2 1999/11/08 11:22:24 shiro Exp $
 */

#include <stk.h>

/* This header file is for other extension modules which need
 * to access uniform vectors in C level.
 */

/* To avoid consuming type tag space, all the uniform vector types
 * shares a common type tag tc_uvector.  Further typing is done by
 * type field of extdata structure.
 */

extern int tc_uvector;          /* uniform vector type tag */

/* common structure */
typedef struct {
    unsigned int type : 8;
    unsigned int length : 24;
} uvector_header;

#define UVECTORP(obj)       TYPEP((obj), tc_uvector)
#define UVECTOR(obj)        ((uvector_header*)EXTDATA(obj))
#define UVECTOR_TYPE(obj)   UVECTOR(obj)->type
#define UVECTOR_LENGTH(obj) UVECTOR(obj)->length

/* subtype tags */
typedef enum {
    S8VECTOR,
    U8VECTOR,
    S16VECTOR,
    U16VECTOR,
    S32VECTOR,
    U32VECTOR,
    S64VECTOR,
    U64VECTOR,
    F32VECTOR,
    F64VECTOR,
    
    UVECTOR_TYPE_MAX
} UvectorType;

#define UVECTOR_TYPEP(obj, type)  (UVECTOR_TYPE(obj) == (unsigned int)(type))

typedef struct {
    char *tagname;
    void (*mark)(SCM);
    void (*display)(SCM, SCM, int);
    SCM (*compare)(SCM, SCM, int);
    SCM (*make_from_length)(SCM, SCM);
    SCM (*make_from_args)(SCM, int);
} uvector_subtype;

/* some generic functions */
extern SCM STk_uvector_p(SCM obj);
extern SCM STk_make_uvector(SCM tag, SCM length, SCM init);
extern SCM STk_uvector(SCM args, int len);

/* s8vector */

typedef struct {
    uvector_header header;
    signed char data[1];
} s8vector_extdata;

extern int tc_s8vector;

#define S8VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, S8VECTOR))
#define S8VECTOR(obj)        ((s8vector_extdata*)EXTDATA(obj))
#define S8VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define S8VECTOR_DATA(obj)   S8VECTOR(obj)->data

extern SCM STk_s8vector_new(int length);
extern SCM STk_s8vector_p(SCM obj);
extern SCM STk_make_s8vector(SCM length, SCM init);
extern SCM STk_s8vector(SCM args, int len);
extern SCM STk_s8vector_length(SCM vec);
extern SCM STk_s8vector_ref(SCM vec, SCM idx);
extern SCM STk_s8vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_s8vector2list(SCM vec);
extern SCM STk_list2s8vector(SCM obj);
extern SCM STk_s8vector_copy(SCM vec);
extern SCM STk_s8vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_s8vector2vector(SCM vec);
extern SCM STk_vector2s8vector(SCM vec);

/* u8vector */

typedef struct {
    uvector_header header;
    unsigned char data[1];
} u8vector_extdata;

#define U8VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, U8VECTOR))
#define U8VECTOR(obj)        ((u8vector_extdata*)EXTDATA(obj))
#define U8VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define U8VECTOR_DATA(obj)   U8VECTOR(obj)->data

extern SCM STk_u8vector_new(int length);
extern SCM STk_u8vector_p(SCM obj);
extern SCM STk_make_u8vector(SCM length, SCM init);
extern SCM STk_u8vector(SCM args, int len);
extern SCM STk_u8vector_length(SCM vec);
extern SCM STk_u8vector_ref(SCM vec, SCM idx);
extern SCM STk_u8vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_u8vector2list(SCM vec);
extern SCM STk_list2u8vector(SCM obj);
extern SCM STk_u8vector_copy(SCM vec);
extern SCM STk_u8vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_u8vector2vector(SCM vec);
extern SCM STk_vector2u8vector(SCM vec);

/* s16vector */

typedef struct {
    uvector_header header;
    signed short data[1];
} s16vector_extdata;

#define S16VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, S16VECTOR))
#define S16VECTOR(obj)        ((s16vector_extdata*)EXTDATA(obj))
#define S16VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define S16VECTOR_DATA(obj)   S16VECTOR(obj)->data

extern SCM STk_s16vector_new(int length);
extern SCM STk_s16vector_p(SCM obj);
extern SCM STk_make_s16vector(SCM length, SCM init);
extern SCM STk_s16vector(SCM args, int len);
extern SCM STk_s16vector_length(SCM vec);
extern SCM STk_s16vector_ref(SCM vec, SCM idx);
extern SCM STk_s16vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_s16vector2list(SCM vec);
extern SCM STk_list2s16vector(SCM obj);
extern SCM STk_s16vector_copy(SCM vec);
extern SCM STk_s16vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_s16vector2vector(SCM vec);
extern SCM STk_vector2s16vector(SCM vec);

/* u16vector */

typedef struct {
    uvector_header header;
    unsigned short data[1];
} u16vector_extdata;

#define U16VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, U16VECTOR))
#define U16VECTOR(obj)        ((u16vector_extdata*)EXTDATA(obj))
#define U16VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define U16VECTOR_DATA(obj)   U16VECTOR(obj)->data

extern SCM STk_u16vector_new(int length);
extern SCM STk_u16vector_p(SCM obj);
extern SCM STk_make_u16vector(SCM length, SCM init);
extern SCM STk_u16vector(SCM args, int len);
extern SCM STk_u16vector_length(SCM vec);
extern SCM STk_u16vector_ref(SCM vec, SCM idx);
extern SCM STk_u16vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_u16vector2list(SCM vec);
extern SCM STk_list2u16vector(SCM obj);
extern SCM STk_u16vector_copy(SCM vec);
extern SCM STk_u16vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_u16vector2vector(SCM vec);
extern SCM STk_vector2u16vector(SCM vec);

/* s32vector */

typedef struct {
    uvector_header header;
    long data[1];
} s32vector_extdata;

#define S32VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, S32VECTOR))
#define S32VECTOR(obj)        ((s32vector_extdata*)EXTDATA(obj))
#define S32VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define S32VECTOR_DATA(obj)   S32VECTOR(obj)->data

extern SCM STk_s32vector_new(int length);
extern SCM STk_s32vector_p(SCM obj);
extern SCM STk_make_s32vector(SCM length, SCM init);
extern SCM STk_s32vector(SCM args, int len);
extern SCM STk_s32vector_length(SCM vec);
extern SCM STk_s32vector_ref(SCM vec, SCM idx);
extern SCM STk_s32vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_s32vector2list(SCM vec);
extern SCM STk_list2s32vector(SCM obj);
extern SCM STk_s32vector_copy(SCM vec);
extern SCM STk_s32vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_s32vector2vector(SCM vec);
extern SCM STk_vector2s32vector(SCM vec);

/* u32vector */

typedef struct {
    uvector_header header;
    unsigned long data[1];
} u32vector_extdata;

#define U32VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, U32VECTOR))
#define U32VECTOR(obj)        ((u32vector_extdata*)EXTDATA(obj))
#define U32VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define U32VECTOR_DATA(obj)   U32VECTOR(obj)->data

extern SCM STk_u32vector_new(int length);
extern SCM STk_u32vector_p(SCM obj);
extern SCM STk_make_u32vector(SCM length, SCM init);
extern SCM STk_u32vector(SCM args, int len);
extern SCM STk_u32vector_length(SCM vec);
extern SCM STk_u32vector_ref(SCM vec, SCM idx);
extern SCM STk_u32vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_u32vector2list(SCM vec);
extern SCM STk_list2u32vector(SCM obj);
extern SCM STk_u32vector_copy(SCM vec);
extern SCM STk_u32vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_u32vector2vector(SCM vec);
extern SCM STk_vector2u32vector(SCM vec);

/* s64vector
 *  This version is a dummy implementation, just to complete
 *  SRFI-4 vector types.  Elements are just SCM objects now.
 */

typedef struct {
    uvector_header header;
    SCM data[1];
} s64vector_extdata;

#define S64VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, S64VECTOR))
#define S64VECTOR(obj)        ((s64vector_extdata*)EXTDATA(obj))
#define S64VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define S64VECTOR_DATA(obj)   S64VECTOR(obj)->data

extern SCM STk_s64vector_new(int length);
extern SCM STk_s64vector_p(SCM obj);
extern SCM STk_make_s64vector(SCM length, SCM init);
extern SCM STk_s64vector(SCM args, int len);
extern SCM STk_s64vector_length(SCM vec);
extern SCM STk_s64vector_ref(SCM vec, SCM idx);
extern SCM STk_s64vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_s64vector2list(SCM vec);
extern SCM STk_list2s64vector(SCM obj);
extern SCM STk_s64vector_copy(SCM vec);
extern SCM STk_s64vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_s64vector2vector(SCM vec);
extern SCM STk_vector2s64vector(SCM vec);

/* u64vector
 *  This version is a dummy implementation, just to complete
 *  SRFI-4 vector types.   Elements are just SCM objects now.
 */

typedef struct {
    uvector_header header;
    SCM data[1];
} u64vector_extdata;

#define U64VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, U64VECTOR))
#define U64VECTOR(obj)        ((u64vector_extdata*)EXTDATA(obj))
#define U64VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define U64VECTOR_DATA(obj)   U64VECTOR(obj)->data

extern SCM STk_u64vector_new(int length);
extern SCM STk_u64vector_p(SCM obj);
extern SCM STk_make_u64vector(SCM length, SCM init);
extern SCM STk_u64vector(SCM args, int len);
extern SCM STk_u64vector_length(SCM vec);
extern SCM STk_u64vector_ref(SCM vec, SCM idx);
extern SCM STk_u64vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_u64vector2list(SCM vec);
extern SCM STk_list2u64vector(SCM obj);
extern SCM STk_u64vector_copy(SCM vec);
extern SCM STk_u64vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_u64vector2vector(SCM vec);
extern SCM STk_vector2u64vector(SCM vec);


/* f32vector */

typedef struct {
    uvector_header header;
    float data[1];
} f32vector_extdata;

#define F32VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, F32VECTOR))
#define F32VECTOR(obj)        ((f32vector_extdata*)EXTDATA(obj))
#define F32VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define F32VECTOR_DATA(obj)   F32VECTOR(obj)->data

extern SCM STk_f32vector_new(int length);
extern SCM STk_f32vector_p(SCM obj);
extern SCM STk_make_f32vector(SCM length, SCM init);
extern SCM STk_f32vector(SCM args, int len);
extern SCM STk_f32vector_length(SCM vec);
extern SCM STk_f32vector_ref(SCM vec, SCM idx);
extern SCM STk_f32vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_f32vector2list(SCM vec);
extern SCM STk_list2f32vector(SCM obj);
extern SCM STk_f32vector_copy(SCM vec);
extern SCM STk_f32vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_f32vector2vector(SCM vec);
extern SCM STk_vector2f32vector(SCM vec);

extern SCM STk_f32vector_scale(SCM vec, SCM factor);
extern SCM STk_f32vector_scale_destructive(SCM vec, SCM factor);
extern SCM STk_f32vector_add(SCM args, int nargs);
extern SCM STk_f32vector_increment(SCM args, int nargs);
extern SCM STk_f32vector_subtract(SCM args, int nargs);
extern SCM STk_f32vector_decrement(SCM args, int nargs);
extern SCM STk_f32vector_dot_product(SCM v1, SCM v2);
extern SCM STk_f32vector_cross_product(SCM v1, SCM v2);

/* f64vector */

typedef struct {
    uvector_header header;
    double data[1];
} f64vector_extdata;

#define F64VECTORP(obj)       (UVECTORP(obj)&&UVECTOR_TYPEP(obj, F64VECTOR))
#define F64VECTOR(obj)        ((f64vector_extdata*)EXTDATA(obj))
#define F64VECTOR_LENGTH(obj) UVECTOR_LENGTH(obj)
#define F64VECTOR_DATA(obj)   F64VECTOR(obj)->data

extern SCM STk_f64vector_new(int length);
extern SCM STk_f64vector_p(SCM obj);
extern SCM STk_make_f64vector(SCM length, SCM init);
extern SCM STk_f64vector(SCM args, int len);
extern SCM STk_f64vector_length(SCM vec);
extern SCM STk_f64vector_ref(SCM vec, SCM idx);
extern SCM STk_f64vector_set(SCM vec, SCM idx, SCM val);
extern SCM STk_f64vector2list(SCM vec);
extern SCM STk_list2f64vector(SCM obj);
extern SCM STk_f64vector_copy(SCM vec);
extern SCM STk_f64vector_copy_destructive(SCM dst, SCM src);
extern SCM STk_f64vector2vector(SCM vec);
extern SCM STk_vector2f64vector(SCM vec);

extern SCM STk_f64vector_scale(SCM vec, SCM factor);
extern SCM STk_f64vector_scale_destructive(SCM vec, SCM factor);
extern SCM STk_f64vector_add(SCM args, int nargs);
extern SCM STk_f64vector_increment(SCM args, int nargs);
extern SCM STk_f64vector_subtract(SCM args, int nargs);
extern SCM STk_f64vector_decrement(SCM args, int nargs);
extern SCM STk_f64vector_dot_product(SCM v1, SCM v2);
extern SCM STk_f64vector_cross_product(SCM v1, SCM v2);


